package com.cmkg.cmkg.process;

import org.apache.spark.mllib.classification.NaiveBayesModel;

import java.io.*;
import java.util.*;

import com.hankcs.hanlp.HanLP;
import com.hankcs.hanlp.seg.Segment;
import com.hankcs.hanlp.seg.common.Term;
import org.apache.spark.SparkConf;
import org.apache.spark.api.java.JavaRDD;
import org.apache.spark.api.java.JavaSparkContext;
import org.apache.spark.mllib.classification.NaiveBayes;
import org.apache.spark.mllib.classification.NaiveBayesModel;
import org.apache.spark.mllib.linalg.Vector;
import org.apache.spark.mllib.linalg.Vectors;
import org.apache.spark.mllib.regression.LabeledPoint;

public class ModelProcess {
    /**
     * ·ÖÀà±êÇ©ºÅºÍÎÊ¾äÄ£°å¶ÔÓ¦±í
     */
    Map<Double, String> questionsPattern;

    /**
     * Spark±´Ò¶Ë¹·ÖÀàÆ÷
     */
    NaiveBayesModel nbModel;

    /**
     * ´ÊÓïºÍÏÂ±êµÄ¶ÔÓ¦±í   == ´Ê»ã±í
     */
    Map<String, Integer> vocabulary;

    /**
     * ¹Ø¼ü×ÖÓëÆä´ÊÐÔµÄmap¼üÖµ¶Ô¼¯ºÏ == ¾ä×Ó³éÏó
     */
    Map<String, String> abstractMap;

    /**
     * Ö¸¶¨ÎÊÌâquestion¼°×ÖµäµÄtxtÄ£°åËùÔÚµÄ¸ùÄ¿Â¼
     */
    String rootDirPath = "D:/HanLP/data";

    /**
     * ·ÖÀàÄ£°åË÷Òý
     */
    int modelIndex = 0;

    public ModelProcess() throws Exception{
        questionsPattern = loadQuestionsPattern();
        vocabulary = loadVocabulary();
        nbModel = loadClassifierModel();
    }


    public ModelProcess(String rootDirPath) throws Exception{
        this.rootDirPath = rootDirPath+'/';
        questionsPattern = loadQuestionsPattern();
        vocabulary = loadVocabulary();
        nbModel = loadClassifierModel();
    }

    public ArrayList<String> analyQuery(String queryString) throws Exception {

        /**
         * ´òÓ¡ÎÊ¾ä
         */
        System.out.println("原始句子："+queryString);
        System.out.println("========HanLP：开始分词========");

        /**
         * ³éÏó¾ä×Ó£¬ÀûÓÃHanLP·Ö´Ê£¬½«¹Ø¼ü×Ö½øÐÐ´ÊÐÔ³éÏó
         */
        String abstr = queryAbstract(queryString);
        System.out.println("抽象句子："+abstr);// nm µÄ µ¼ÑÝ ÊÇ Ë­

        /**
         * ½«³éÏóµÄ¾ä×ÓÓësparkÑµÁ·¼¯ÖÐµÄÄ£°å½øÐÐÆ¥Åä£¬ÄÃµ½¾ä×Ó¶ÔÓ¦µÄÄ£°å
         */
        String strPatt = queryClassify(abstr);
        System.out.println("句子套用模版结果："+strPatt); // nm ÖÆ×÷ µ¼ÑÝÁÐ±í

        /**
         * Ä£°å»¹Ô­³É¾ä×Ó£¬´ËÊ±ÎÊÌâÒÑ×ª»»ÎªÎÒÃÇÊìÏ¤µÄ²Ù×÷
         */
        String finalPattern = queryExtenstion(strPatt);
        System.out.println("原始句子替换成系统可识别的结果"+finalPattern);// µ«¶¡ÃÜÂë ÖÆ×÷ µ¼ÑÝÁÐ±í

        ArrayList<String> resultList = new ArrayList<String>();
        resultList.add(String.valueOf(modelIndex));
        String[] finalPattArray = finalPattern.split(" ");
        for(String word : finalPattArray)
            resultList.add(word);
        return resultList;
    }

    public String queryAbstract(String querySentence) {

        // ¾ä×Ó³éÏó»¯
        Segment segment = HanLP.newSegment().enableCustomDictionary(true);
        List<Term> terms = segment.seg(querySentence);
        String abstractQuery = "";
        abstractMap = new HashMap<String, String>();
        for (Term term : terms) {
            String word = term.word;
            String termStr = term.toString();
            System.out.println(termStr);
            if (termStr.contains("nm")) {        //nm 病
                abstractQuery += "nm ";
                abstractMap.put("nm", word);
            }else if (termStr.contains("nhm")) { //nhm 药
                abstractQuery += "nhm ";
                abstractMap.put("nhm", word);
            }
            else {
                abstractQuery += word + " ";
            }
        }
        System.out.println("========HanLP·Ö´Ê½áÊø========");
        return abstractQuery;
    }

    public  String queryExtenstion(String queryPattern) {
        // ¾ä×Ó»¹Ô­
        Set<String> set = abstractMap.keySet();
        for(String key : set) {
            /**
             * Èç¹û¾ä×ÓÄ£°åÖÐº¬ÓÐ³éÏóµÄ´ÊÐÔ
             */
            if (queryPattern.contains(key)) {
                /**
                 * ÔòÌæ»»³éÏó´ÊÐÔÎª¾ßÌåµÄÖµ
                 */
                String value = abstractMap.get(key);
                queryPattern = queryPattern.replace(key, value);
            }
        }
        String extendedQuery = queryPattern;
        /**
         * µ±Ç°¾ä×Ó´¦ÀíÍê£¬³éÏómapÇå¿ÕÊÍ·Å¿Õ¼ä²¢ÖÃ¿Õ£¬µÈ´ýÏÂÒ»¸ö¾ä×ÓµÄ´¦Àí
         */
        abstractMap.clear();
        abstractMap = null;
        return extendedQuery;
    }


    /**
     * ¼ÓÔØ´Ê»ã±í == ¹Ø¼üÌØÕ÷ == ÓëHanLP·Ö´ÊºóµÄµ¥´Ê½øÐÐÆ¥Åä
     * @return
     */
    public Map<String, Integer> loadVocabulary() {
        Map<String, Integer> vocabulary = new HashMap<String, Integer>();
        File file = new File(rootDirPath + "question/vocabulary.txt");
        BufferedReader br = null;
        try {
            br = new BufferedReader(new FileReader(file));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
        String line;
        try {
            while ((line = br.readLine()) != null) {

                System.out.println(line);
                String[] tokens = line.split("：");
                System.out.println(tokens[0]);
                int index = Integer.parseInt(tokens[0]);
                String word = tokens[1];
                vocabulary.put(word, index);
            }
        }catch (NumberFormatException e) {
            e.printStackTrace();
        }catch (IOException e) {
            e.printStackTrace();
        }
        return vocabulary;
    }

    /**
     * ¼ÓÔØÎÄ¼þ£¬²¢¶ÁÈ¡ÄÚÈÝ·µ»Ø
     * @param filename
     * @return
     * @throws IOException
     */
    public String loadFile(String filename) throws IOException {
        File file = new File(rootDirPath + filename);
        BufferedReader br = new BufferedReader(new FileReader(file));
        String content = "";
        String line;
        while ((line = br.readLine()) != null) {
            /**
             * ÎÄ±¾µÄ»»ÐÐ·ûÔÝ¶¨ÓÃ"`"´úÌæ
             */
            content += line + "`";
        }
        /**
         * ¹Ø±Õ×ÊÔ´
         */
        br.close();
        return content;
    }

    /**
     * ¾ä×Ó·Ö´ÊºóÓë´Ê»ã±í½øÐÐkeyÆ¥Åä×ª»»ÎªdoubleÏòÁ¿Êý×é
     * @param sentence
     * @return
     * @throws Exception
     */
    public double[] sentenceToArrays(String sentence) throws Exception {

        double[] vector = new double[vocabulary.size()];
        /**
         * Ä£°å¶ÔÕÕ´Ê»ã±íµÄ´óÐ¡½øÐÐ³õÊ¼»¯£¬È«²¿Îª0.0
         */
        for(int i = 0; i < vocabulary.size(); i++) {
            vector[i] = 0;
        }

        /**
         * HanLP·Ö´Ê£¬ÄÃ·Ö´ÊµÄ½á¹ûºÍ´Ê»ã±íÀïÃæµÄ¹Ø¼üÌØÕ÷½øÐÐÆ¥Åä
         */
        Segment segment = HanLP.newSegment();
        List<Term> terms = segment.seg(sentence);
        for(Term term : terms) {
            String word = term.word;
//            System.out.println(word);
            /**
             * Èç¹ûÃüÖÐ£¬0.0 ¸ÄÎª 1.0
             */
            if(vocabulary.containsKey(word)) {
                int index = vocabulary.get(word);
                vector[index] = 1;
            }
        }

        return vector;
    }

    /**
     * SparkÆÓËØ±´Ò¶Ë¹(naiveBayes)
     * ¶ÔÌØ¶¨µÄÄ£°å½øÐÐ¼ÓÔØ²¢·ÖÀà
     * ÓûÁË½âSparkÆÓËØ±´Ò¶Ë¹£¬¿É²Î¿¼µØÖ·£ºhttps://blog.csdn.net/appleyk/article/details/80348912
     * @return
     * @throws Exception
     */
    public NaiveBayesModel loadClassifierModel() throws Exception {

        /**
         * Éú³ÉSpark¶ÔÏó
         * Ò»¡¢Spark³ÌÐòÊÇÍ¨¹ýSparkContext·¢²¼µ½Spark¼¯ÈºµÄ
         * Spark³ÌÐòµÄÔËÐÐ¶¼ÊÇÔÚSparkContextÎªºËÐÄµÄµ÷¶ÈÆ÷µÄÖ¸»ÓÏÂ½øÐÐµÄ
         * Spark³ÌÐòµÄ½áÊøÊÇÒÔSparkContext½áÊø×÷Îª½áÊø
         * JavaSparkContext¶ÔÏóÓÃÀ´´´½¨SparkµÄºËÐÄRDDµÄ
         * ×¢Òâ£ºµÚÒ»¸öRDD,Ò»¶¨ÊÇÓÉSparkContextÀ´´´½¨µÄ
         *
         * ¶þ¡¢SparkContextµÄÖ÷¹¹ÔìÆ÷²ÎÊýÎª SparkConf
         * SparkConf±ØÐëÉèÖÃappnameºÍmaster£¬·ñÔò»á±¨´í
         * spark.master   ÓÃÓÚÉèÖÃ²¿ÊðÄ£Ê½
         * local[*] == ±¾µØÔËÐÐÄ£Ê½[Ò²¿ÉÒÔÊÇ¼¯ÈºµÄÐÎÊ½]£¬Èç¹ûÐèÒª¶à¸öÏß³ÌÖ´ÐÐ£¬¿ÉÒÔÉèÖÃÎªlocal[2],±íÊ¾2¸öÏß³Ì £¬*±íÊ¾¶à¸ö
         * spark.app.name ÓÃÓÚÖ¸¶¨Ó¦ÓÃµÄ³ÌÐòÃû³Æ  ==
         */

        SparkConf conf = new SparkConf().setAppName("NaiveBayesTest").setMaster("local[*]");
        JavaSparkContext sc = new JavaSparkContext(conf);

        /**
         * ÑµÁ·¼¯Éú³É
         * labeled point ÊÇÒ»¸ö¾Ö²¿ÏòÁ¿£¬ÒªÃ´ÊÇÃÜ¼¯ÐÍµÄÒªÃ´ÊÇÏ¡ÊèÐÍµÄ
         * ÓÃÒ»¸ölabel/response½øÐÐ¹ØÁª¡£ÔÚMLlibÀï£¬labeled points ±»ÓÃÀ´¼à¶½Ñ§Ï°Ëã·¨
         * ÎÒÃÇÊ¹ÓÃÒ»¸ödoubleÊýÀ´´æ´¢Ò»¸ölabel£¬Òò´ËÎÒÃÇÄÜ¹»Ê¹ÓÃlabeled points½øÐÐ»Ø¹éºÍ·ÖÀà
         */
        List<LabeledPoint> train_list = new LinkedList<LabeledPoint>();
        String[] sentences = null;


        /**
         * ÌÇÄò²¡µÄÖÎÁÆ·½·¨ÊÇÊ²Ã´
         */
        String scoreQuestions = loadFile("question/therapeuticmethod.txt");
        sentences = scoreQuestions.split("`");
        for (String sentence : sentences) {
            double[] array = sentenceToArrays(sentence);
            LabeledPoint train_one = new LabeledPoint(0.0, Vectors.dense(array));
            train_list.add(train_one);
        }

        /**
         * ÌÇÄò²¡µÄÖ¢×´ÊÇÊ²Ã´
         */
        String timeQuestions = loadFile("question/Symptom.txt");
        sentences = timeQuestions.split("`");
        for (String sentence : sentences) {
            double[] array = sentenceToArrays(sentence);
            LabeledPoint train_one = new LabeledPoint(1.0, Vectors.dense(array));
            train_list.add(train_one);
        }


        /**
         * Ò©µÄÓÃ·¨ÊÇÊ²Ã´
         */
        String styleQuestions = loadFile("question/Operation.txt");
        sentences = styleQuestions.split("`");
        for (String sentence : sentences) {
            double[] array = sentenceToArrays(sentence);
            LabeledPoint train_one = new LabeledPoint(2.0, Vectors.dense(array));
            train_list.add(train_one);
        }


        /**
         * Ò©µÄ×÷ÓÃÊÇÊ²Ã´
         */
        String storyQuestions = loadFile("question/Function.txt");
        sentences = storyQuestions.split("`");
        for (String sentence : sentences) {
            double[] array = sentenceToArrays(sentence);
            LabeledPoint train_one = new LabeledPoint(3.0, Vectors.dense(array));
            train_list.add(train_one);
        }


        /**
         * SPARKµÄºËÐÄÊÇRDD(µ¯ÐÔ·Ö²¼Ê½Êý¾Ý¼¯)
         * SparkÊÇScalaÐ´µÄ,JavaRDD¾ÍÊÇSparkÎªJavaÐ´µÄÒ»Ì×API
         * JavaSparkContext sc = new JavaSparkContext(sparkConf);    //¶ÔÓ¦JavaRDD
         * SparkContext	    sc = new SparkContext(sparkConf)    ;    //¶ÔÓ¦RDD
         */
        JavaRDD<LabeledPoint> trainingRDD = sc.parallelize(train_list);
        NaiveBayesModel nb_model = NaiveBayes.train(trainingRDD.rdd());

        /**
         * ¼ÇµÃ¹Ø±Õ×ÊÔ´
         */
        sc.close();

        /**
         * ·µ»Ø±´Ò¶Ë¹·ÖÀàÆ÷
         */
        return nb_model;

    }

    /**
     * ¼ÓÔØÎÊÌâÄ£°å == ·ÖÀàÆ÷±êÇ©
     * @return
     */
    public  Map<Double, String> loadQuestionsPattern() {
        Map<Double, String> questionsPattern = new HashMap<Double, String>();
        File file = new File(rootDirPath + "question/question_classification.txt");
        BufferedReader br = null;
        try {
            br = new BufferedReader(new FileReader(file));
        } catch (FileNotFoundException e1) {
            e1.printStackTrace();
        }
        String line;
        try {
            while ((line = br.readLine()) != null) {
                String[] tokens = line.split(":");
                double index = Double.valueOf(tokens[0]);
                String pattern = tokens[1];
                questionsPattern.put(index, pattern);
            }
        } catch (NumberFormatException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return questionsPattern;
    }

    /**
     * ±´Ò¶Ë¹·ÖÀàÆ÷·ÖÀàµÄ½á¹û£¬ÄÃµ½Æ¥ÅäµÄ·ÖÀà±êÇ©ºÅ£¬²¢¸ù¾Ý±êÇ©ºÅ·µ»ØÎÊÌâµÄÄ£°å
     * @param sentence
     * @return
     * @throws Exception
     */
    public  String queryClassify(String sentence) throws Exception {
        System.out.print(sentence);
        double[] testArray = sentenceToArrays(sentence);
        for(int i=0;i<testArray.length;i++) System.out.print(""+testArray[i]+"");
        Vector v = Vectors.dense(testArray);

        /**
         * ¶ÔÊý¾Ý½øÐÐÔ¤²âpredict
         * ¾ä×ÓÄ£°åÔÚ spark±´Ò¶Ë¹·ÖÀàÆ÷ÖÐµÄË÷Òý¡¾Î»ÖÃ¡¿
         * ¸ù¾Ý´Ê»ãÊ¹ÓÃµÄÆµÂÊÍÆ¶Ï³ö¾ä×Ó¶ÔÓ¦ÄÄÒ»¸öÄ£°å
         */
        double index = nbModel.predict(v);
        modelIndex = (int)index;
        System.out.println("the model index is " + index);
//		Vector vRes = nbModel.predictProbabilities(v);
//		System.out.println("ÎÊÌâÄ£°å·ÖÀà¡¾0¡¿¸ÅÂÊ£º"+vRes.toArray()[0]);
//		System.out.println("ÎÊÌâÄ£°å·ÖÀà¡¾13¡¿¸ÅÂÊ£º"+vRes.toArray()[13]);
        return questionsPattern.get(index);
    }

    public static void main(String[] agrs) throws Exception {
        //ÐèÒªµ÷ÓÃÉÏÃæµÄº¯ÊýÊµÏÖ¹¦ÄÜ£¬´ý½â¾ö
    }
}
